package java.awt.geom;

import java.io.Serializable;
import java.awt.Rectangle;
import sun.awt.geom.Curve;
import java.util.Arrays;
import java.awt.Shape;

public abstract class CubicCurve2D implements Shape, Cloneable
{
    protected CubicCurve2D() {
    }
    
    public abstract double getX1();
    
    public abstract double getY1();
    
    public abstract Point2D getP1();
    
    public abstract double getCtrlX1();
    
    public abstract double getCtrlY1();
    
    public abstract Point2D getCtrlP1();
    
    public abstract double getCtrlX2();
    
    public abstract double getCtrlY2();
    
    public abstract Point2D getCtrlP2();
    
    public abstract double getX2();
    
    public abstract double getY2();
    
    public abstract Point2D getP2();
    
    public abstract void setCurve(double p0, double p1, double p2, double p3, double p4, double p5, double p6, double p7);
    
    public void setCurve(double[] array, int n) {
        this.setCurve(array[n + 0], array[n + 1], array[n + 2], array[n + 3], array[n + 4], array[n + 5], array[n + 6], array[n + 7]);
    }
    
    public void setCurve(Point2D point2D, Point2D point2D2, Point2D point2D3, Point2D point2D4) {
        this.setCurve(point2D.getX(), point2D.getY(), point2D2.getX(), point2D2.getY(), point2D3.getX(), point2D3.getY(), point2D4.getX(), point2D4.getY());
    }
    
    public void setCurve(Point2D[] array, int n) {
        this.setCurve(array[n + 0].getX(), array[n + 0].getY(), array[n + 1].getX(), array[n + 1].getY(), array[n + 2].getX(), array[n + 2].getY(), array[n + 3].getX(), array[n + 3].getY());
    }
    
    public void setCurve(CubicCurve2D cubicCurve2D) {
        this.setCurve(cubicCurve2D.getX1(), cubicCurve2D.getY1(), cubicCurve2D.getCtrlX1(), cubicCurve2D.getCtrlY1(), cubicCurve2D.getCtrlX2(), cubicCurve2D.getCtrlY2(), cubicCurve2D.getX2(), cubicCurve2D.getY2());
    }
    
    public static double getFlatnessSq(double n, double n2, double n3, double n4, double n5, double n6, double n7, double n8) {
        return Math.max(Line2D.ptSegDistSq(n, n2, n7, n8, n3, n4), Line2D.ptSegDistSq(n, n2, n7, n8, n5, n6));
    }
    
    public static double getFlatness(double n, double n2, double n3, double n4, double n5, double n6, double n7, double n8) {
        return Math.sqrt(getFlatnessSq(n, n2, n3, n4, n5, n6, n7, n8));
    }
    
    public static double getFlatnessSq(double[] array, int n) {
        return getFlatnessSq(array[n + 0], array[n + 1], array[n + 2], array[n + 3], array[n + 4], array[n + 5], array[n + 6], array[n + 7]);
    }
    
    public static double getFlatness(double[] array, int n) {
        return getFlatness(array[n + 0], array[n + 1], array[n + 2], array[n + 3], array[n + 4], array[n + 5], array[n + 6], array[n + 7]);
    }
    
    public double getFlatnessSq() {
        return getFlatnessSq(this.getX1(), this.getY1(), this.getCtrlX1(), this.getCtrlY1(), this.getCtrlX2(), this.getCtrlY2(), this.getX2(), this.getY2());
    }
    
    public double getFlatness() {
        return getFlatness(this.getX1(), this.getY1(), this.getCtrlX1(), this.getCtrlY1(), this.getCtrlX2(), this.getCtrlY2(), this.getX2(), this.getY2());
    }
    
    public void subdivide(CubicCurve2D cubicCurve2D, CubicCurve2D cubicCurve2D2) {
        subdivide(this, cubicCurve2D, cubicCurve2D2);
    }
    
    public static void subdivide(CubicCurve2D cubicCurve2D, CubicCurve2D cubicCurve2D2, CubicCurve2D cubicCurve2D3) {
        double x1 = cubicCurve2D.getX1();
        double y1 = cubicCurve2D.getY1();
        double ctrlX1 = cubicCurve2D.getCtrlX1();
        double ctrlY1 = cubicCurve2D.getCtrlY1();
        double ctrlX2 = cubicCurve2D.getCtrlX2();
        double ctrlY2 = cubicCurve2D.getCtrlY2();
        double x2 = cubicCurve2D.getX2();
        double y2 = cubicCurve2D.getY2();
        double n = (ctrlX1 + ctrlX2) / 2.0;
        double n2 = (ctrlY1 + ctrlY2) / 2.0;
        double n3 = (x1 + ctrlX1) / 2.0;
        double n4 = (y1 + ctrlY1) / 2.0;
        double n5 = (x2 + ctrlX2) / 2.0;
        double n6 = (y2 + ctrlY2) / 2.0;
        double n7 = (n3 + n) / 2.0;
        double n8 = (n4 + n2) / 2.0;
        double n9 = (n5 + n) / 2.0;
        double n10 = (n6 + n2) / 2.0;
        double n11 = (n7 + n9) / 2.0;
        double n12 = (n8 + n10) / 2.0;
        if (cubicCurve2D2 != null) {
            cubicCurve2D2.setCurve(x1, y1, n3, n4, n7, n8, n11, n12);
        }
        if (cubicCurve2D3 != null) {
            cubicCurve2D3.setCurve(n11, n12, n9, n10, n5, n6, x2, y2);
        }
    }
    
    public static void subdivide(double[] array, int n, double[] array2, int n2, double[] array3, int n3) {
        double n4 = array[n + 0];
        double n5 = array[n + 1];
        double n6 = array[n + 2];
        double n7 = array[n + 3];
        double n8 = array[n + 4];
        double n9 = array[n + 5];
        double n10 = array[n + 6];
        double n11 = array[n + 7];
        if (array2 != null) {
            array2[n2 + 0] = n4;
            array2[n2 + 1] = n5;
        }
        if (array3 != null) {
            array3[n3 + 6] = n10;
            array3[n3 + 7] = n11;
        }
        double n12 = (n4 + n6) / 2.0;
        double n13 = (n5 + n7) / 2.0;
        double n14 = (n10 + n8) / 2.0;
        double n15 = (n11 + n9) / 2.0;
        double n16 = (n6 + n8) / 2.0;
        double n17 = (n7 + n9) / 2.0;
        double n18 = (n12 + n16) / 2.0;
        double n19 = (n13 + n17) / 2.0;
        double n20 = (n14 + n16) / 2.0;
        double n21 = (n15 + n17) / 2.0;
        double n22 = (n18 + n20) / 2.0;
        double n23 = (n19 + n21) / 2.0;
        if (array2 != null) {
            array2[n2 + 2] = n12;
            array2[n2 + 3] = n13;
            array2[n2 + 4] = n18;
            array2[n2 + 5] = n19;
            array2[n2 + 6] = n22;
            array2[n2 + 7] = n23;
        }
        if (array3 != null) {
            array3[n3 + 0] = n22;
            array3[n3 + 1] = n23;
            array3[n3 + 2] = n20;
            array3[n3 + 3] = n21;
            array3[n3 + 4] = n14;
            array3[n3 + 5] = n15;
        }
    }
    
    public static int solveCubic(double[] array) {
        return solveCubic(array, array);
    }
    
    public static int solveCubic(double[] array, double[] array2) {
        double n = array[3];
        if (n == 0.0) {
            return QuadCurve2D.solveQuadratic(array, array2);
        }
        double n2 = array[2] / n;
        double n3 = array[1] / n;
        double n4 = array[0] / n;
        double n5 = n2 * n2;
        double n6 = 0.3333333333333333 * (-0.3333333333333333 * n5 + n3);
        double n7 = 0.5 * (0.07407407407407407 * n2 * n5 - 0.3333333333333333 * n2 * n3 + n4);
        double n8 = n6 * n6 * n6;
        double a = n7 * n7 + n8;
        double a2 = 0.3333333333333333 * n2;
        int fixRoots;
        if (a < 0.0) {
            double a4 = 0.3333333333333333 * Math.acos(-n7 / Math.sqrt(-n8));
            double n11 = 2.0 * Math.sqrt(-n6);
            if (array2 == array) {
                array = Arrays.copyOf(array, 4);
            }
            array2[0] = n11 * Math.cos(a4);
            array2[1] = -n11 * Math.cos(a4 + 1.0471975511965976);
            array2[2] = -n11 * Math.cos(a4 - 1.0471975511965976);
            fixRoots = 3;
            for (int i = 0; i < fixRoots; ++i) {
                int n12 = i;
                array2[n12] -= a2;
            }
        }
        else {
            double sqrt = Math.sqrt(a);
            double cbrt = Math.cbrt(sqrt - n7);
            double n9 = -Math.cbrt(sqrt + n7);
            double a3 = cbrt + n9;
            fixRoots = 1;
            double n10 = 1.2E9 * Math.ulp(Math.abs(a3) + Math.abs(a2));
            if (iszero(a, n10) || within(cbrt, n9, n10)) {
                if (array2 == array) {
                    array = Arrays.copyOf(array, 4);
                }
                array2[1] = -(a3 / 2.0) - a2;
                fixRoots = 2;
            }
            array2[0] = a3 - a2;
        }
        if (fixRoots > 1) {
            fixRoots = fixRoots(array, array2, fixRoots);
        }
        if (fixRoots > 2 && (array2[2] == array2[1] || array2[2] == array2[0])) {
            --fixRoots;
        }
        if (fixRoots > 1 && array2[1] == array2[0]) {
            array2[1] = array2[--fixRoots];
        }
        return fixRoots;
    }
    
    private static int fixRoots(double[] array, double[] a, int toIndex) {
        double[] array2 = { array[1], 2.0 * array[2], 3.0 * array[3] };
        int solveQuadratic = QuadCurve2D.solveQuadratic(array2, array2);
        if (solveQuadratic == 2 && array2[0] == array2[1]) {
            --solveQuadratic;
        }
        if (solveQuadratic == 2 && array2[0] > array2[1]) {
            double n = array2[0];
            array2[0] = array2[1];
            array2[1] = n;
        }
        if (toIndex == 3) {
            double rootUpperBound = getRootUpperBound(array);
            double n5 = -rootUpperBound;
            Arrays.sort(a, 0, toIndex);
            if (solveQuadratic == 2) {
                a[0] = refineRootWithHint(array, n5, array2[0], a[0]);
                a[1] = refineRootWithHint(array, array2[0], array2[1], a[1]);
                a[2] = refineRootWithHint(array, array2[1], rootUpperBound, a[2]);
                return 3;
            }
            if (solveQuadratic == 1) {
                double n6 = array[3];
                double n7 = -n6;
                double n8 = array2[0];
                double solveEqn2 = solveEqn(array, 3, n8);
                if (oppositeSigns(n7, solveEqn2)) {
                    a[0] = bisectRootWithHint(array, n5, n8, a[0]);
                }
                else if (oppositeSigns(solveEqn2, n6)) {
                    a[0] = bisectRootWithHint(array, n8, rootUpperBound, a[2]);
                }
                else {
                    a[0] = n8;
                }
            }
            else if (solveQuadratic == 0) {
                a[0] = bisectRootWithHint(array, n5, rootUpperBound, a[1]);
            }
        }
        else if (toIndex == 2 && solveQuadratic == 2) {
            double n2 = a[0];
            double n3 = a[1];
            double n4 = array2[0];
            double d = array2[1];
            d = ((Math.abs(n4 - n2) > Math.abs(d - n2)) ? n4 : d);
            double solveEqn = solveEqn(array, 3, d);
            if (iszero(solveEqn, 1.0E7 * Math.ulp(d))) {
                a[1] = ((Math.abs(solveEqn(array, 3, n3)) < Math.abs(solveEqn)) ? n3 : d);
                return 2;
            }
        }
        return 1;
    }
    
    private static double refineRootWithHint(double[] array, double n, double n2, double d) {
        if (!inInterval(d, n, n2)) {
            return d;
        }
        double[] array2 = { array[1], 2.0 * array[2], 3.0 * array[3] };
        int i = 0;
        while (i < 3) {
            double solveEqn = solveEqn(array2, 2, d);
            double solveEqn2 = solveEqn(array, 3, d);
            double n3 = d + -(solveEqn2 / solveEqn);
            if (solveEqn != 0.0 && solveEqn2 != 0.0) {
                if (d == n3) {
                    break;
                }
                d = n3;
                ++i;
            }
            else {
                break;
            }
        }
        if (within(d, d, 1000.0 * Math.ulp(d)) && inInterval(d, n, n2)) {
            return d;
        }
        return d;
    }
    
    private static double bisectRootWithHint(double[] array, double n, double n2, double n3) {
        double min = Math.min(Math.abs(n3 - n) / 64.0, 0.0625);
        double min2 = Math.min(Math.abs(n3 - n2) / 64.0, 0.0625);
        double n4 = n3 - min;
        double n5 = n3 + min2;
        double n6 = solveEqn(array, 3, n4);
        double n7 = solveEqn(array, 3, n5);
        while (oppositeSigns(n6, n7)) {
            if (n4 >= n5) {
                return n4;
            }
            n = n4;
            n2 = n5;
            min /= 64.0;
            min2 /= 64.0;
            n4 = n3 - min;
            n5 = n3 + min2;
            n6 = solveEqn(array, 3, n4);
            n7 = solveEqn(array, 3, n5);
        }
        if (n6 == 0.0) {
            return n4;
        }
        if (n7 == 0.0) {
            return n5;
        }
        return bisectRoot(array, n, n2);
    }
    
    private static double bisectRoot(double[] array, double n, double n2) {
        double solveEqn = solveEqn(array, 3, n);
        double n3 = n + (n2 - n) / 2.0;
        while (n3 != n && n3 != n2) {
            double solveEqn2 = solveEqn(array, 3, n3);
            if (solveEqn2 == 0.0) {
                return n3;
            }
            if (oppositeSigns(solveEqn, solveEqn2)) {
                n2 = n3;
            }
            else {
                solveEqn = solveEqn2;
                n = n3;
            }
            n3 = n + (n2 - n) / 2.0;
        }
        return n3;
    }
    
    private static boolean inInterval(double n, double n2, double n3) {
        return n2 <= n && n <= n3;
    }
    
    private static boolean within(double n, double n2, double n3) {
        double n4 = n2 - n;
        return n4 <= n3 && n4 >= -n3;
    }
    
    private static boolean iszero(double n, double n2) {
        return within(n, 0.0, n2);
    }
    
    private static boolean oppositeSigns(double n, double n2) {
        return (n < 0.0 && n2 > 0.0) || (n > 0.0 && n2 < 0.0);
    }
    
    private static double solveEqn(double[] array, int n, double n2) {
        double n3 = array[n];
        while (--n >= 0) {
            n3 = n3 * n2 + array[n];
        }
        return n3;
    }
    
    private static double getRootUpperBound(double[] array) {
        double d = 1.0 + Math.max(Math.max(Math.abs(array[2]), Math.abs(array[1])), Math.abs(array[0])) / Math.abs(array[3]);
        return d + (Math.ulp(d) + 1.0);
    }
    
    @Override
    public boolean contains(double n, double n2) {
        if (n * 0.0 + n2 * 0.0 != 0.0) {
            return false;
        }
        double x1 = this.getX1();
        double y1 = this.getY1();
        double x2 = this.getX2();
        double y2 = this.getY2();
        return (Curve.pointCrossingsForLine(n, n2, x1, y1, x2, y2) + Curve.pointCrossingsForCubic(n, n2, x1, y1, this.getCtrlX1(), this.getCtrlY1(), this.getCtrlX2(), this.getCtrlY2(), x2, y2, 0) & 0x1) == 0x1;
    }
    
    @Override
    public boolean contains(Point2D point2D) {
        return this.contains(point2D.getX(), point2D.getY());
    }
    
    @Override
    public boolean intersects(double n, double n2, double n3, double n4) {
        return n3 > 0.0 && n4 > 0.0 && this.rectCrossings(n, n2, n3, n4) != 0;
    }
    
    @Override
    public boolean intersects(Rectangle2D rectangle2D) {
        return this.intersects(rectangle2D.getX(), rectangle2D.getY(), rectangle2D.getWidth(), rectangle2D.getHeight());
    }
    
    @Override
    public boolean contains(double n, double n2, double n3, double n4) {
        if (n3 <= 0.0 || n4 <= 0.0) {
            return false;
        }
        int rectCrossings = this.rectCrossings(n, n2, n3, n4);
        return rectCrossings != 0 && rectCrossings != Integer.MIN_VALUE;
    }
    
    private int rectCrossings(double n, double n2, double n3, double n4) {
        int rectCrossingsForLine = 0;
        if (this.getX1() != this.getX2() || this.getY1() != this.getY2()) {
            rectCrossingsForLine = Curve.rectCrossingsForLine(rectCrossingsForLine, n, n2, n + n3, n2 + n4, this.getX1(), this.getY1(), this.getX2(), this.getY2());
            if (rectCrossingsForLine == Integer.MIN_VALUE) {
                return rectCrossingsForLine;
            }
        }
        return Curve.rectCrossingsForCubic(rectCrossingsForLine, n, n2, n + n3, n2 + n4, this.getX2(), this.getY2(), this.getCtrlX2(), this.getCtrlY2(), this.getCtrlX1(), this.getCtrlY1(), this.getX1(), this.getY1(), 0);
    }
    
    @Override
    public boolean contains(Rectangle2D rectangle2D) {
        return this.contains(rectangle2D.getX(), rectangle2D.getY(), rectangle2D.getWidth(), rectangle2D.getHeight());
    }
    
    @Override
    public Rectangle getBounds() {
        return this.getBounds2D().getBounds();
    }
    
    @Override
    public PathIterator getPathIterator(AffineTransform affineTransform) {
        return new CubicIterator(this, affineTransform);
    }
    
    @Override
    public PathIterator getPathIterator(AffineTransform affineTransform, double n) {
        return new FlatteningPathIterator(this.getPathIterator(affineTransform), n);
    }
    
    public Object clone() {
        try {
            return super.clone();
        }
        catch (CloneNotSupportedException cause) {
            throw new InternalError(cause);
        }
    }
    
    public static class Float extends CubicCurve2D implements Serializable
    {
        public float x1;
        public float y1;
        public float ctrlx1;
        public float ctrly1;
        public float ctrlx2;
        public float ctrly2;
        public float x2;
        public float y2;
        private static final long serialVersionUID = -1272015596714244385L;
        
        public Float() {
        }
        
        public Float(float n, float n2, float n3, float n4, float n5, float n6, float n7, float n8) {
            this.setCurve(n, n2, n3, n4, n5, n6, n7, n8);
        }
        
        @Override
        public double getX1() {
            return (double)this.x1;
        }
        
        @Override
        public double getY1() {
            return (double)this.y1;
        }
        
        @Override
        public Point2D getP1() {
            return new Point2D.Float(this.x1, this.y1);
        }
        
        @Override
        public double getCtrlX1() {
            return (double)this.ctrlx1;
        }
        
        @Override
        public double getCtrlY1() {
            return (double)this.ctrly1;
        }
        
        @Override
        public Point2D getCtrlP1() {
            return new Point2D.Float(this.ctrlx1, this.ctrly1);
        }
        
        @Override
        public double getCtrlX2() {
            return (double)this.ctrlx2;
        }
        
        @Override
        public double getCtrlY2() {
            return (double)this.ctrly2;
        }
        
        @Override
        public Point2D getCtrlP2() {
            return new Point2D.Float(this.ctrlx2, this.ctrly2);
        }
        
        @Override
        public double getX2() {
            return (double)this.x2;
        }
        
        @Override
        public double getY2() {
            return (double)this.y2;
        }
        
        @Override
        public Point2D getP2() {
            return new Point2D.Float(this.x2, this.y2);
        }
        
        @Override
        public void setCurve(double n, double n2, double n3, double n4, double n5, double n6, double n7, double n8) {
            this.x1 = (float)n;
            this.y1 = (float)n2;
            this.ctrlx1 = (float)n3;
            this.ctrly1 = (float)n4;
            this.ctrlx2 = (float)n5;
            this.ctrly2 = (float)n6;
            this.x2 = (float)n7;
            this.y2 = (float)n8;
        }
        
        public void setCurve(float x1, float y1, float ctrlx1, float ctrly1, float ctrlx2, float ctrly2, float x2, float y2) {
            this.x1 = x1;
            this.y1 = y1;
            this.ctrlx1 = ctrlx1;
            this.ctrly1 = ctrly1;
            this.ctrlx2 = ctrlx2;
            this.ctrly2 = ctrly2;
            this.x2 = x2;
            this.y2 = y2;
        }
        
        @Override
        public Rectangle2D getBounds2D() {
            float min = Math.min(Math.min(this.x1, this.x2), Math.min(this.ctrlx1, this.ctrlx2));
            float min2 = Math.min(Math.min(this.y1, this.y2), Math.min(this.ctrly1, this.ctrly2));
            return new Rectangle2D.Float(min, min2, Math.max(Math.max(this.x1, this.x2), Math.max(this.ctrlx1, this.ctrlx2)) - min, Math.max(Math.max(this.y1, this.y2), Math.max(this.ctrly1, this.ctrly2)) - min2);
        }
    }
    
    public static class Double extends CubicCurve2D implements Serializable
    {
        public double x1;
        public double y1;
        public double ctrlx1;
        public double ctrly1;
        public double ctrlx2;
        public double ctrly2;
        public double x2;
        public double y2;
        private static final long serialVersionUID = -4202960122839707295L;
        
        public Double() {
        }
        
        public Double(double n, double n2, double n3, double n4, double n5, double n6, double n7, double n8) {
            this.setCurve(n, n2, n3, n4, n5, n6, n7, n8);
        }
        
        @Override
        public double getX1() {
            return this.x1;
        }
        
        @Override
        public double getY1() {
            return this.y1;
        }
        
        @Override
        public Point2D getP1() {
            return new Point2D.Double(this.x1, this.y1);
        }
        
        @Override
        public double getCtrlX1() {
            return this.ctrlx1;
        }
        
        @Override
        public double getCtrlY1() {
            return this.ctrly1;
        }
        
        @Override
        public Point2D getCtrlP1() {
            return new Point2D.Double(this.ctrlx1, this.ctrly1);
        }
        
        @Override
        public double getCtrlX2() {
            return this.ctrlx2;
        }
        
        @Override
        public double getCtrlY2() {
            return this.ctrly2;
        }
        
        @Override
        public Point2D getCtrlP2() {
            return new Point2D.Double(this.ctrlx2, this.ctrly2);
        }
        
        @Override
        public double getX2() {
            return this.x2;
        }
        
        @Override
        public double getY2() {
            return this.y2;
        }
        
        @Override
        public Point2D getP2() {
            return new Point2D.Double(this.x2, this.y2);
        }
        
        @Override
        public void setCurve(double x1, double y1, double ctrlx1, double ctrly1, double ctrlx2, double ctrly2, double x2, double y2) {
            this.x1 = x1;
            this.y1 = y1;
            this.ctrlx1 = ctrlx1;
            this.ctrly1 = ctrly1;
            this.ctrlx2 = ctrlx2;
            this.ctrly2 = ctrly2;
            this.x2 = x2;
            this.y2 = y2;
        }
        
        @Override
        public Rectangle2D getBounds2D() {
            double min = Math.min(Math.min(this.x1, this.x2), Math.min(this.ctrlx1, this.ctrlx2));
            double min2 = Math.min(Math.min(this.y1, this.y2), Math.min(this.ctrly1, this.ctrly2));
            return new Rectangle2D.Double(min, min2, Math.max(Math.max(this.x1, this.x2), Math.max(this.ctrlx1, this.ctrlx2)) - min, Math.max(Math.max(this.y1, this.y2), Math.max(this.ctrly1, this.ctrly2)) - min2);
        }
    }
}
